REBOL [
	Title: "REBOL Mezzanine Functions: File"
	Version: 2.7.6
	Rights: "Copyright REBOL Technologies 2008. All rights reserved."
	Home: http://www.rebol.com
	Date: 14-Mar-2008

	; You are free to use, modify, and distribute this file as long as the
	; above header, copyright, and this entire comment remains intact.
	; This software is provided "as is" without warranties of any kind.
	; In no event shall REBOL Technologies or source contributors be liable
	; for any damages of any kind, even if advised of the possibility of such
	; damage. See license for more information.

	; Please help us to improve this software by contributing changes and
	; fixes. See http://www.rebol.com/support.html for details.
	; You are free to use, modify, and distribute this software with any
	; REBOL Technologies products as long as the above header, copyright,
	; and this comment remain intact. This software is provided "as is"
	; and without warranties of any kind. In no event shall the owners or
	; contributors be liable for any damages of any kind, even if advised
	; of the possibility of such damage. See license for more information.

	; Please help us to improve this software by contributing changes and
	; fixes via http://www.rebol.com/feedback.html - Thanks!
]

info?: func [
	{Returns information about a file or url.}
	[catch]
	target [file! url!]
][
	throw-on-error [
		target: make port! target
		query target		; update data
	]
	either none? target/status [
		none
	][
		make object! [
			size: target/size date: target/date type: target/status
		]
	]
]

exists?: func [
	{Determines if a file or URL exists.}
	target [file! url!]
	/local ret
][
	either error? try [
		target: make port! target
		query target
	][ false ][found? target/status]
]

size?: func [
	{Returns the size of a file or URL's contents.}
	[catch]
	target [file! url!]
	/local info
][
	info: throw-on-error[ info? target ]
	either none? info [none][info/size]
]

modified?: func [
	{Returns the last modified date of a file or URL.}
	[catch]
	target [file! url!]
	/local info
][
	info: throw-on-error[ info? target ]
	either none? info [none][info/date]
]

dir?: func [
	{Returns TRUE if a file or URL is a directory.}
	[catch]
	target [file! url!]
	/local info
][
	info: throw-on-error[ info? target ]
	either none? info [false][info/type = 'directory]
]

echo: func [
	{Copies console output to a file.}
	[catch]
	target [file! none! logic!]
][
	if port? system/ports/echo [
		close system/ports/echo
		system/ports/echo: none
	]
	if target = true [target: %rebol-echo.txt]
	if file? target [
		system/ports/echo: throw-on-error[ open/write/new/direct target ]
	]
]

what-dir: func [
	"Prints the active directory path"
][
	copy system/script/path
]

change-dir: func [
	"Changes the active directory path."
	[catch]
	dir [file!] {new directory path}
	/local slash full ret
][
	dir: clean-path dir
	if (last dir) <> #"/" [append dir #"/"]
	if (first dir) = #"/" [
		slash: next dir
		full: remove parse system/script/path "/"
		while [all [not empty? slash (first slash) = #"/" not empty? full]] [
			insert slash first full
			remove full
			slash: find/tail slash "/"
			if (empty? full) or (empty? slash) [clear slash break]
		]
	]
	either throw-on-error[ dir? dir ] [
		system/script/path: dir
	][
		throw make error! reduce ['access 'cannot-open dir]
	]
	system/script/path
]

list-dir: func [
	"Prints a multi-column sorted listing of a directory."
	[catch]
	dir [file! url! unset!] {Directory to list or nothing}
	/local file max s
][
	dir: throw-on-error [
		either value? 'dir [    ; how to take nothing as an arg
			; RAMBO #3723
			read dirize dir
		][
			read %.
		]
	]
	dir: sort dir
	max: 0
	foreach file dir [
		if (length? file) > max [max: length? file]
	]
	max: max + 2
	s: make string! 0
	foreach file dir [
		append s file
		while [((length? s) // max) <> 0][append s " "]
		if (length? s) > 60 [
			print s
			clear s
		]
	]
	if length? s [print s]
]

rename: func [
	"Renames a file to a new name."
	[catch]
	old [file! url!]            {path to the old file}
	new [file! url! string!]    {new name (not a path)}
	/local p dir blk
][
	dir: open first p: split-path clean-path old
	blk: copy dir
	either not none? blk: find blk second p [       ; find the entry for target
		throw-on-error [ change skip dir ((index? blk) - 1) new ]
	][
		throw make error! reduce ['access 'no-rename old]
	]
	close dir
]

clean-path: func [
	"Cleans-up '.' and '..' in path; returns the cleaned path."
	target [file! url!]] [
	if url? target [return target]
	if find [%/ %/. %/..] target [return %/] ; workaround for request-file
	get-modes target 'full-path
]

split-path: func [
	"Splits a file or URL path. Returns a block containing path and target."
	target [file! url!]
	/local dir pos
][
	parse/all target [
		[#"/" | 1 2 #"." opt #"/"] end (dir: dirize target) |
		pos: any [thru #"/" [end | pos:]] (
			all [empty? dir: copy/part target at head target index? pos dir: %./]
			all [find [%. %..] pos: to file! pos insert tail pos #"/"]
		)
	]
	reduce [dir pos]
]

delete: func [
	"Deletes the specified file(s)."
	[catch]
	target [file! url!] {the file to delete}
	/any                    {allow wild cards}
	/local p dir blk file ret err
] [
	dir: throw-on-error [ open first p: split-path clean-path target ]
	error? set/any 'err try [
		either any [                ; if wildcards enabled
			blk: copy dir           ; get block of names
			foreach file blk [
				ret: find/match/any file second p
				if none? ret [ret: file]
				either tail? ret [
					remove dir
				][
					dir: next dir
				]
			]
		] [
			blk: copy dir
			either not none? blk: find blk second p [   ; find the entry for target
				remove skip dir (index? blk) - 1
			] [
				make error! reduce ['access 'no-delete target] 
			]
		]
	]
	attempt [close dir]
	either error? get/any 'err [throw err][get/any 'err]
]

dirize: func [
	{Returns a copy of the path turned into a directory.}
	path [file! string! url!]
][
	either #"/" <> pick path length? path [join path #"/"] [copy path]
]

undirize: func [
	{Returns a copy of the path with any trailing "/" removed.}
	path [file! string! url!]
][
	path: copy path
	if #"/" = pick path length? path [clear back tail path]
	path
]

make-dir: func [
	"Creates the specified directory. No error if already exists."
	[catch]
	path [file! url!]
	/deep "Create subdirectories too"
	/local dirs end created
][
	if empty? path [return path]
	if slash <> last path [path: dirize path]
	if exists? path [
		if dir? path [return path]
		return make error! reduce ['access 'cannot-open path]
	]
	if any [not deep url? path] [
		close throw-on-error [open/new path]
		return path
	]
	; Scan reverse looking for first existing dir:
	path: copy path
	dirs: copy []
	while [
		all [
			not empty? path
			not exists? path
			remove back tail path ; trailing slash
		]
	][
		end: any [find/last/tail path slash path]
		insert dirs copy end
		clear end
	]
	; Create directories forward:
	created: copy []
	foreach dir dirs [
		path: either empty? path [dir][path/:dir]
		append path slash
		if error? try [make-dir path] [
			foreach dir created [attempt [delete dir]]
			return make error! reduce ['access 'cannot-open path]
		]
		insert created path
	]
	path
]

delete-dir: func [
	"Deletes a directory including all files and subdirectories."
	dir [file! url!]
	/local files
][
	if all [
		dir? dir
		dir: dirize dir
		attempt [files: load dir]
	][
		foreach file files [delete-dir dir/:file]
	]
	attempt [delete dir]
]

suffix?: func [
	"Return the suffix (ext) of a filename or url, else NONE."
	path [any-string!]
][
	if all [
		path: find/last path #"."
		not find path #"/"
	][to-file path]
]

in-dir: func [
	"Evaluate a block while in a directory."
	[throw]
	dir [file!] "Directory to change to (changed back after)"
	block [block!] "Block to evaluate"
	/local old-dir
][
	old-dir: what-dir
	change-dir dir
	also do block change-dir old-dir
]

to-relative-file: func [
	"Returns the relative portion of a file if in a subdirectory, or the original if not."
	file [file! string!] "File to check (local if string!)"
	/no-copy "Don't copy, just reference"
	/as-rebol "Convert to REBOL-style filename if not"
	/as-local "Convert to local-style filename if not"
	/local tmp
] [
	either string? file [ ; Local file (Note: to-local-file drops trailing /)
		if tmp: find/match file to-local-file what-dir [file: next tmp]
		if as-rebol [file: to-rebol-file file  no-copy: true]
	] [
		file: any [find/match file what-dir  file]
		if as-local [file: to-local-file file  no-copy: true]
	]
	unless no-copy [file: copy file]
	file
]
