REBOL [
	Title: "REBOL Mezzanine Functions: Series"
	Version: 2.7.7
	Rights: "Copyright REBOL Technologies 2009. All rights reserved."
	Home: http://www.rebol.com
	Date: 30-Dec-2009

	; You are free to use, modify, and distribute this file as long as the
	; above header, copyright, and this entire comment remains intact.
	; This software is provided "as is" without warranties of any kind.
	; In no event shall REBOL Technologies or source contributors be liable
	; for any damages of any kind, even if advised of the possibility of such
	; damage. See license for more information.

	; Please help us to improve this software by contributing changes and
	; fixes. See http://www.rebol.com/support.html for details.

	; Portions of this code are licensed from the R2/Forward project.
	; You are free to use, modify, and distribute this software with any
	; REBOL Technologies products as long as the above header, copyright,
	; and this comment remain intact. This software is provided "as is"
	; and without warranties of any kind. In no event shall the owners or
	; contributors be liable for any damages of any kind, even if advised
	; of the possibility of such damage. See license for more information.

	; Please help us to improve this software by contributing changes and
	; fixes via http://www.rebol.com/feedback.html - Thanks!
]

offset?: func [
	{Returns the offset between two series positions.}
	series1 [series!]
	series2 [series!]
][
	subtract index? series2 index? series1
]

found?: func [
	{Returns TRUE if value is not NONE.}
	value
][
	not none? :value
]

append: func [
    {Appends a value to the tail of a series and returns the series head.}
    series [series! port!]
    value
    /only {Appends a block value as a block}
][
	head either only [
		insert/only tail series :value
	][
		insert tail series :value
	]
]

repend: func [
    {Appends a reduced value to a series and returns the series head.}
    series [series! port!]
    value
    /only {Appends a block value as a block}
][
	head either only [
		insert/only tail series reduce :value
	][
		insert tail series reduce :value
	]
]

join: func [
	{Concatenates values.}
	value {Base value}
	rest {Value or block of values}
][
	value: either series? :value [copy value] [form :value]
	repend value :rest
]

rejoin: func [
	{Reduces and joins a block of values.}
	block [block!] {Values to reduce and join}
	;/with "separator"
][
	if empty? block: reduce block [return block]
	append either series? first block [copy first block][
		form first block] next block
]

ajoin: func [
	"Reduces and joins a block of values into a new string."
	[throw]
	block [block!]
][
	make string! reduce block
]
; Note: Native in R3 without the reduce overhead.

reform: func [
	{Forms a reduced block and returns a string.}
	value {Value to reduce and form}
	;/with "separator"
][
	form reduce :value
]

remold: func [
	{Molds a reduced block and returns a string.}
	value {Value to reduce and mold}
][
	mold reduce :value
]

charset: func [
	{Makes a bitset of chars for the parse function.}
	chars [string! block!]
][
	make bitset! chars
]

array: func [
	"Makes and initializes a series of a given size."
	[catch]
	size [integer! block!] "Size or block of sizes for each dimension"
	/initial "Specify an initial value for all elements"
	value "Initial value (will be called each time if a function)"
	/local block rest
][
	if block? size [
		if tail? rest: next size [rest: none]
		unless integer? set/any 'size pick size 1 [
			throw-error 'script 'expect-arg reduce ['array 'size type? get/any 'size]
		]
	]
	block: make block! size
	case [
		block? rest [
			loop size [block: insert/only block array/initial rest :value]
		]
		series? :value [
			loop size [block: insert/only block copy/deep value]
		]
		any-function? :value [ ; So value can be a thunk :)
			loop size [block: insert/only block value] ; Called every time
		]
		insert/dup block value size
	]
	head block
]
; Note: Function values are evaluated at end of block to ensure safety.

replace: func [
	"Replaces the search value with the replace value within the target series."
	target  [series!] "Series that is being modified"
	search  "Value to be replaced"
	replace "Value to replace with (will be called each time if a function)"
	/all "Replace all occurrences"  ;!!! Note ALL is redefined in here!
	/case "Case-sensitive replacement"  ;!!! Note CASE is redefined in here!
	/tail "Return target after the last replacement position"  ;!!! Note TAIL is redefined in here!
	/local save-target len value pos do-break
][
	save-target: target
	; If target is a string but search is not, make search a string (except for bitset).
	; If target is a bitset, or a block and search is not a block, len = 1
	len: system/words/case [
		bitset? :search  1
		any-string? target [
			if any [not any-string? :search tag? :search] [search: form :search]
			length? :search
		]
		any-block? :search [length? :search]
		true  1		
	]
	; /all and /case checked before the while, /tail after
	do-break: unless all [:break] ; Will be none if not /all, a noop
	while pick [
		[pos: find target :search]
		[pos: find/case target :search]
	] not case [
		(value: replace pos) ; The replace argument can be a function
		target: change/part pos :value len
		do-break
	]
	either tail [target] [save-target]
]
; Note: Code injection vulnerability with get-word! parameters. Needs APPLY.
; An /any refinement is waiting for FIND/any to work in R3.

move: func [
	"Move a value or span of values in a series."
	[catch]
	source [series!] "Source series"
	offset [integer!] "Offset to move by, or index to move to"
	/part "Move part of a series"
	length [integer!] "The length of the part to move"
	/skip "Treat the series as records of fixed size" ;; SKIP redefined
	size [integer!] "Size of each record"
	/to "Move to an index relative to the head of the series" ;; TO redefined
][
	unless length [length: 1]
	if skip [
		if 1 > size [throw-error 'script 'out-of-range size]
		offset: offset * size
		length: length * size
	]
	part: copy/part source length
	remove/part source length
	insert either to [at head source offset] [
		system/words/skip source offset
	] part
]
; Note: This is the best you can do without overlap and aliasing issues.
; The R3 version is included to prevent it from being reinvented, badly.
; Suggested by a discussion in AltME in Nov-2007.

swap: func [
	"Swaps elements of a series. (Modifies)"
	series1 [series!]
	series2 [series!]
][
	unless any [empty? series1 empty? series2] [
		poke series1 1 also pick series2 1 poke series2 1 pick series1 1
	]
	series1
]
; Note: Native (action!) in R3.

extract: func [
	"Extracts a value from a series at regular intervals."
	[catch]
	series [series!]
	width [integer!] "Size of each entry (the skip)"
	/index "Extract from an offset position"
	pos "The position" [number! logic! block!]
	/default "Use a default value instead of none"
	value "The value to use (will be called each time if a function)"
	/into "Insert into a buffer instead (returns position after insert)"
	output [series!] "The buffer series (modified)"
	/local len
][
	if zero? width [return any [output make series 0]]  ; To avoid an infinite loop
	len: either positive? width [  ; Length to preallocate
		divide length? series width  ; Forward loop, use length
	][
		divide index? series negate width  ; Backward loop, use position
	]
	unless index [pos: 1]
	either block? pos [
		if empty? pos [return any [output make series 0]] ; Shortcut return
		parse pos [some [number! | logic! | set pos skip (
			throw-error 'script 'expect-set reduce [[number! logic!] type? get/any 'pos]
		)]]
		unless into [output: make series len * length? pos]
		if all [not default any-string? output] [value: copy ""]
		; R2 PARSE doesn't work well for binary!, so spoof a string!.
		if binary? series [series: as-string series]
		forskip series width [forall pos [output: insert/only output any [pick series pos/1 value]]]
	][
		unless into [output: make series len]
		if all [not default any-string? output] [value: copy ""]
		; R2 PARSE doesn't work well for binary!, so spoof a string!.
		if binary? series [series: as-string series]
		forskip series width [output: insert/only output any [pick series pos value]]
	]
	either into [output] [head output]
]
; Note: Function values are evaluated at end of block to ensure safety.

alter: func [
	"If a value is not found in a series, append it; otherwise, remove it. Returns true if added. (Modifies)"
	series [series! port!]
	value
	/case "Case-sensitive comparison"
][
	found? unless remove (
		either case [find/case series :value] [find series :value]
	) [append series :value]
]

first+: func [
	{Return FIRST of series, and increment the series index.}
	[catch]
	'word [word! paren!] "Word must be a series."  ; paren! added for R2
][
	; Workaround for R3 change in lit-word! parameters with paren! arguments
	if paren? :word [set/any 'word do :word] 
	throw-on-error [also pick get word 1 set word next get word]
]
; Note: Native in R3.

take: func [
	"Copies and removes from series. (Modifies)"
	[catch]
	value [series! port! none!]
	/part "Limits to a given length or position"
	length [number! series! port! pair!]
	/last "Take it from the tail end"
][
	if value [
		either part [
			case [
				pair? length [
					unless image? value [
						throw-error 'script 'invalid-part length
					]
					last: none
				]
				any [series? length port? length] [
					either same? head value head length [
						length: subtract index? length index? value
					][
						throw-error 'script 'invalid-part reduce [length]
					]
				]
			]
			if last [
				length: negate length
				value: tail value
			]
			also copy/part value length remove/part value length
		][
			also pick either last [
				value: back tail value
			] [value] 1 remove value
		]
	]
]
; Note: Native (action!) in R3.

; MAP-EACH with set-words, best datatype! support and /into (ideal full version)
map-each: func [
	"Evaluates a block for each value(s) in a series and returns them as a block."
	[throw catch]
	'word [word! block!] "Word or block of words to set each time (local)"
	data [block!] "The series to traverse"
	body [block!] "Block to evaluate each time"
	/into "Collect into a given series, rather than a new block"
	output [any-block! any-string!] "The series to output to" ; Not image!
	/local init len x
][
	; Shortcut return for empty data
	if empty? data [return any [output make block! 0]]
	; BIND/copy word and body
	word: either block? word [
		if empty? word [throw make error! [script invalid-arg []]]
		copy/deep word  ; /deep because word is rebound before errors checked
	] [reduce [word]]
	word: use word reduce [word]
	body: bind/copy body first word
	; Build init code
	init: none
	parse word [any [word! | x: set-word! (
		unless init [init: make block! 4]
		; Add [x: at data index] to init, and remove from word
		insert insert insert tail init first x [at data] index? x
		remove x
	) :x | x: skip (
		throw make error! reduce ['script 'expect-set [word! set-word!] type? first x]
	)]]
	len: length? word ; Can be zero now (for advanced code tricks)
	; Create the output series if not specified
	unless into [output: make block! divide length? data max 1 len]
	; Process the data (which is not empty at this point)
	until [ ; Note: output: insert/only output needed for list! output
		set word data  do init
		unless unset? set/any 'x do body [output: insert/only output :x]
		tail? data: skip data len
	]
	; Return the output and clean up memory references
	also either into [output] [head output] (
		set [word data body output init x] none
	)
]
; Note: This is pretty fast by R2 mezzanine loop standards, native in R3.

collect: func [
	"Evaluates a block, storing values via KEEP function, and returns block of collected values."
	body [block!] "Block to evaluate"
	/into "Insert into a buffer instead (returns position after insert)"
	output [series!] "The buffer series (modified)"
][
	unless output [output: make block! 16]
	do make function! [keep] copy/deep body make function! [value /only] copy/deep [
		output: either only [insert/only output :value] [insert output :value]
		:value
	]
	either into [output] [head output]
]
; Note: R3 version uses the new FUNC, which does the copy/deep already.
; R3 version based on a discussion with Gregg and Gabriele in AltME.


; Character/string encoding functions

;use [
;	ascii utf8+1 utf8+2 utf8+3 utf8+4 utf8+5 utf8rest
;	utf8+1-latin1 utf8rest-latin1
;] [
;ascii: charset [#"^(00)" - #"^(7F)"]
;utf8+1: charset [#"^(C0)" - #"^(DF)"]
;utf8+2: charset [#"^(E0)" - #"^(EF)"]
;utf8+3: charset [#"^(F0)" - #"^(F7)"]
;utf8+4: charset [#"^(F8)" - #"^(FB)"]
;utf8+5: charset [#"^(FC)" - #"^(FD)"]
;utf8rest: charset [#"^(80)" - #"^(BF)"]
;utf8+1-latin1: charset [#"^(C0)" - #"^(C3)"]
;utf8rest-latin1: charset [#"^(80)" - #"^(83)"] ; For bad UTF-8

ascii?: funct [
	"Returns TRUE if value or string is in ASCII character range (below 128)."
	value [any-string! char! integer!]
] compose [
	ascii: (charset [#"^(00)" - #"^(7F)"])
	either any-string? value [parse/all/case value [any ascii]] [value < 128]
]
; Note: Native in R3.

latin1?: funct [
	"Returns TRUE if value or string is in Latin-1 character range (below 256)."
	value [any-string! char! integer!]
] compose [ ; Decodes UTF-8 if binary!
	ascii: (charset [#"^(00)" - #"^(7F)"])
	utf8+1-latin1: (charset [#"^(C0)" - #"^(C3)"])
	utf8rest: (charset [#"^(80)" - #"^(BF)"])
	utf8rest-latin1: (charset [#"^(80)" - #"^(83)"]) ; For bad UTF-8
	switch/default type?/word value [
		integer! [value < 256]
		binary! [parse/all/case value [any [
			ascii | utf8+1-latin1 utf8rest | ; Minimized Latin-1
			#{E0} utf8rest-latin1 utf8rest | ; Bad 3-byte Latin-1
			#{F080} utf8rest-latin1 utf8rest | ; Bad 4-byte Latin-1
			#{F88080} utf8rest-latin1 utf8rest | ; Bad 5-byte Latin-1
			#{FC808080} utf8rest-latin1 utf8rest ; Bad 6-byte Latin-1
		]]]
	] [true] ; R2 has Latin-1 chars and strings
]
; Note: Native in R3. Bug of accepting non-minimized UTF-8 in R3 too, for now.

utf?: func [
	"Returns the UTF encoding from the BOM (byte order marker): + for BE; - for LE."
	data [binary!]
][
	parse/all/case data [
		#{EF BB BF} (return 8) |
		#{00 00 FE FF} (return 32) |
		#{FF FE 00 00} (return -32) |
		#{FE FF} (return 16) |
		#{FF FE} (return -16) |
		(return 0) ; No BOM or an unsupported BOM
	] ; Note: R3 only supports these BOMs.
]
; Note: Native in R3.

;]

deline: funct [
	"Converts string terminators to standard format, e.g. CRLF to LF. (Modifies)"
	string [any-string!]
	/lines "Convert to block of lines (does not modify)"
] compose [
	linechar: (complement charset crlf)
	also case [
		not lines [ ; Convert the string in place
			; Change the newlines from any platform-specific line ending
			; R2 PARSE doesn't work well for binary!, so spoof a string!
			parse/all/case either binary? string [as-string string] [string] [
				any [to cr a: cr opt lf b: (
					b: change/part a lf b
				) :b] to end
			]
			string
		] ; Otherwise generate a block of lines
		empty? string [copy []] ; Shortcut return for empty string
		'else [
			output: make block! divide length? string 50 ; Seems OK
			; R2 PARSE doesn't work well for binary!, so spoof a string!
			if binary? string [string: as-string string]
			parse/all/case string [any [
				; Lines with stuff in them
				copy a some linechar (output: insert output a) [crlf | cr | lf | end]
				|
				; Empty lines
				[crlf | cr | lf] (output: insert output copy "")
			]]
			head output
		]
	] set [string output a b] none ; Cleanup after return
] ; /lines needs testing with final line ending once it works in R3.
; Note: Native in R3.

enline: funct [
	"Converts standard string terminators to current OS format, e.g. LF to CRLF. (Modifies)"
	series [any-string! block!]
	/with "Specifies alternate line termination."
	end-of-line [char! string!]
] compose [
	; The platform-specific line ending
	platform-line: (switch/default system/version/4 [
		3 15 [crlf] 2 [to-string either system/version/5 < 4 [cr] [lf]]
	] [to-string lf]) ; Don't know about Amiga (1) and Tao Elate (27)
	switch type?/word end-of-line [
		none! [end-of-line: platform-line]
		char! [end-of-line: to-string end-of-line]
	]
	either block? series [
		; First pass: Precalculate the length of the output string
		len: 0 len-eol: length? end-of-line
		foreach s series [len: len + len-eol + length? s]
		; Second pass: Build and fill the output string
		output: make string! len
		foreach s series [insert insert tail output s end-of-line]
		; Return and clean up
		also output set [series output] none
	] [
		unless end-of-line = "^/" [
			; Change the newlines to the platform-specific line ending
			; R2 PARSE doesn't work well for binary!, so spoof a string!
			parse/all/case either binary? series [as-string series] [series] [
				any [to lf a: lf (a: change/part a end-of-line 1) :a] to end
			]
		]
		; Return and clean up
		also series set [series a] none
	]
] ; block! needs testing with final line ending once it works in R3.
; Note: Native in R3.
